
/**
 *
 * Tennix! SDL Port
 * Copyright (C) 2003, 2007, 2008, 2009 Thomas Perl <thp@thpinfo.com>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA  02110-1301, USA.
 *
 **/

#include "tennix.h"
#include "sound.h"
#include "archive.h"
#include "graphics.h"

#include <stdarg.h>
#include <stdlib.h>
#include <assert.h>

static Sound* sounds;

static char* samples[] = {
    "racket1.wav",
    "racket2.wav",
    "rain.wav",
    "ground1.wav",
    "ground2.wav",
    "audience.wav",
    "applause.wav",
    "out.wav",
    "background.wav",
    "mouseover.wav",
    "mouseclick.wav",

#ifdef HAVE_VOICE_FILES
    "to.wav",
    "all.wav",
    "love-in.wav",
    "love-out.wav",
    "fifteen-in.wav",
    "fifteen-out.wav",
    "thirty-in.wav",
    "thirty-out.wav",
    "forty-in.wav",
    "forty-out.wav",
    "deuce.wav",
    "advantage-player-one.wav",
    "advantage-player-two.wav",

    "zero-in.wav",
    "zero-out.wav",
    "one-in.wav",
    "one-out.wav",
    "two-in.wav",
    "two-out.wav",
    "three-in.wav",
    "three-out.wav",
    "four-in.wav",
    "four-out.wav",
    "five-in.wav",
    "five-out.wav",
    "six-in.wav",
    "six-out.wav",
    "seven-in.wav",
    "seven-out.wav",

    "in-the-first-set.wav",
    "in-the-second-set.wav",
    "in-the-third-set.wav",
    "in-the-fourth-set.wav",
    "in-the-fifth-set.wav",

    "quit-it1.wav",
    "quit-it2.wav",
    "quit-it3.wav",
    "quit-it4.wav",

    "new-game1.wav",
    "new-game2.wav",
    "new-game3.wav",
    "new-game4.wav",
    "new-game5.wav",
    "new-game6.wav",

    "lets-go1.wav",
    "lets-go2.wav",
    "lets-go3.wav",
    "lets-go4.wav",
#endif
};

static sound_id voice_queue[VOICE_QUEUE_MAX];
static int voice_queue_size = 0;
static int voice_queue_position = 0;

int voice_finished_flag = 1;

void init_sound() {
    int i;
    char *d;
    Mix_Chunk* data;
    TennixArchive* tnxar;

    if( Mix_OpenAudio( 44100, AUDIO_S16SYS, 2, 1024) < 0) {
        fprintf( stderr, "Error initializing SDL_mixer: %s\n", Mix_GetError());
    }

    sounds = (Sound*)calloc(SOUND_MAX, sizeof(Sound));

    draw_button(40, (HEIGHT-40)/2, WIDTH-80, 40, 100, 100, 100, 1);
    store_screen();
    updatescr();

    tnxar = tnxar_open(ARCHIVE_FILE);
    if (tnxar == NULL) {
        /* not found in cwd - try installed... */
        tnxar = tnxar_open(ARCHIVE_FILE_INSTALLED);
        assert(tnxar != NULL);
    }

    for( i=0; i<SOUND_MAX; i++) {
        if (tnxar_set_current_filename(tnxar, samples[i]) != 0) {
            d = tnxar_read_current(tnxar);
            data = Mix_LoadWAV_RW(SDL_RWFromMem(d, tnxar_size_current(tnxar)), 0);
            free(d);
        } else {
            fprintf(stderr, "Cannot find %s\n", samples[i]);
            exit(EXIT_FAILURE);
        }
        if( !data) {
            fprintf( stderr, "Error: %s\n", SDL_GetError());
            continue;
        }

        sounds[i].data = data;

        draw_button(40, (HEIGHT-40)/2, (WIDTH-80)*(GR_COUNT+i)/(SOUND_MAX+GR_COUNT), 40, 100, 250, 100, 0);
        rectangle(40+BUTTON_BORDER, (HEIGHT-40)/2+20, (WIDTH-80)*(GR_COUNT+i)/(SOUND_MAX+GR_COUNT)-2*BUTTON_BORDER, 10, 170, 250, 170);
        updatescr();
    }

    tnxar_close(tnxar);

    clear_screen();
    store_screen();
    updatescr();

    /* for Voice Queue processing */
    Mix_ChannelFinished(voice_channel_finished);
}

void play_sample_n(sound_id id, int n)
{
    if (id >= SOUND_MAX) {
        fprintf(stderr, "Cannot play sound #%d.\n", id);
        return;
    }

    if (n == -2) {
        Mix_FadeInChannel(CHANNEL_BY_ID(id), sounds[id].data, -1, FADE_IN_MS);
    } else {
        Mix_PlayChannel(CHANNEL_BY_ID(id), sounds[id].data, n);
    }

    /* Audience stops clapping when ball is served */
    if (id >= SOUND_RACKET_FIRST && id <= SOUND_RACKET_LAST) {
        stop_sample(SOUND_APPLAUSE);
    }
}

void stop_sample(sound_id id)
{
    Mix_FadeOutChannel(CHANNEL_BY_ID(id), FADE_OUT_MS);
}

void pan_sample(sound_id id, float position)
{
    if (position == 0.5) {
        Mix_SetPanning(CHANNEL_BY_ID(id), 255, 255);
    }
    else {
        Mix_SetPanning(CHANNEL_BY_ID(id), 255*(1.0-position), 255*(position));
    }
}

void voice_clear()
{
    voice_queue_size = 0;
    Mix_HaltChannel(CHANNEL_VOICE);
}

void voice_enqueue(sound_id id)
{
    if (voice_queue_size < VOICE_QUEUE_MAX) {
        voice_queue[voice_queue_size++] = id;
    } else {
        fprintf(stderr, "Voice queue overflow. Skipping: %d\n", id);
    }
}

void voice_say()
{
    if (voice_queue_size > 0) {
        voice_queue_position = 0;
        voice_finished_flag = 0;
        Mix_PlayChannel(CHANNEL_VOICE, sounds[voice_queue[voice_queue_position]].data, 0);
        voice_queue_position++;
    }
}

void voice_say_list(int n, ...)
{
    va_list ap;
    int i;
    sound_id id;

    voice_clear();

    va_start(ap, n);
    for (i=0; i<n; i++) {
        id = (sound_id)va_arg(ap, int);
        voice_enqueue(id);
    }
    va_end(ap);
    voice_say();
}

void voice_channel_finished(int channel)
{
    if (channel == CHANNEL_VOICE) {
        if (voice_queue_position < voice_queue_size) {
            Mix_PlayChannel(CHANNEL_VOICE, sounds[voice_queue[voice_queue_position]].data, 0);
            voice_queue_position++;
        } else {
            voice_finished_flag = 1;
            voice_queue_position = 0;
        }
    }
}

